"""Contains the log handler to catch log"""

from __future__ import annotations

import contextlib
import logging
import re
from functools import partial
from logging import LogRecord
from typing import Callable

from .runbot_env import RunbotExcludeWarning, RunbotStepConfig

_logger = logging.getLogger("odoo_runbot.log_filter")
# 7-bit C1 ANSI sequences
ansi_escape = re.compile(
    r"""
    \x1B  # ESC
    (?:   # 7-bit C1 Fe (except CSI)
        [@-Z\\-_]
    |     # or [ for CSI, followed by a control sequence
        \[
        [0-?]*  # Parameter bytes
        [ -/]*  # Intermediate bytes
        [@-~]   # Final byte
    )
""",
    re.VERBOSE,
)

__all__ = [
    "ExcludeWarningFilter",
    "RunbotWarningWatcherHandler",
    "get_handler",
    "start_warning_log_watcher",
]
LOGGING_FORMAT = "%(name)s:::%(message)s"


class RunbotWarningWatcherHandler(logging.Handler):
    filters: list[ExcludeWarningFilter]

    def __init__(self) -> None:
        super().__init__(logging.WARNING)
        self.set_name(type(self).__name__)
        self.setFormatter(logging.Formatter(LOGGING_FORMAT))
        self.catch_all_filter = self._get_catch_all_filter()
        self._log_emit_filter: list[ExcludeWarningFilter] = []

    def _get_catch_all_filter(self) -> ExcludeWarningFilter:
        return ExcludeWarningFilter(
            RunbotExcludeWarning(
                name="Runbot Warning catch all no filter",
                min_match=0,
                max_match=0,  # No warning is accepted
                regex=r".*",  # Every thing it's a match
            ),
        )

    def record_matchers(self) -> list[ExcludeWarningFilter]:
        return list(self._log_emit_filter)

    def emit(self, record: LogRecord) -> None:
        "Do nothing, only here to store logging."
        if record.name.startswith("odoo_runbot"):
            return
        matched = False
        formated_msg = self.format(record)
        for emit_filter in self._log_emit_filter:
            matched = matched or emit_filter.is_a_match(record, formated_msg)
        if not matched:
            self.catch_all_filter.is_a_match(record, formated_msg)

    def add_warnings(self, warnings_to_filter: list[RunbotExcludeWarning]) -> None:
        self._log_emit_filter.extend([ExcludeWarningFilter(warn_rule) for warn_rule in warnings_to_filter])

    def remove_warnings(
        self,
        warnings_to_filter: list[RunbotExcludeWarning] | None = None,
    ) -> list[ExcludeWarningFilter]:
        result: list[ExcludeWarningFilter] = []
        if not warnings_to_filter:
            result = self._log_emit_filter[:]
            self._log_emit_filter = []

        for _filter in self._log_emit_filter[:]:
            if _filter.exclude in warnings_to_filter:
                result.append(_filter)
                self._log_emit_filter.remove(_filter)

        if not self._log_emit_filter:
            result.append(self.catch_all_filter)
            self.catch_all_filter = self._get_catch_all_filter()

        return result

    def handle(self, record: LogRecord) -> None:
        if record.levelno > logging.WARNING:
            return None
        return super().handle(record)


def get_handler() -> RunbotWarningWatcherHandler | None:
    for h in logging.root.handlers:
        if isinstance(h, RunbotWarningWatcherHandler):
            return h
    return None


class ExcludeWarningFilter:
    def __init__(self, exclude: RunbotExcludeWarning) -> None:
        self.log_match: list[logging.LogRecord] = []
        self.regex = re.compile(exclude.regex, re.IGNORECASE)
        self.exclude = exclude
        _logger.info(
            "Init filter '%s' match [%s] for logger %s, between [%s, %s]",
            self.exclude.name,
            str(self.regex),
            self.exclude.logger,
            self.exclude.min_match,
            self.exclude.max_match,
        )

    def is_a_match(self, original_record: logging.LogRecord, formated_msg: str) -> bool:
        _logger.debug("OHOHOHO a new log to filter: %s", formated_msg)
        logger_name, log_msg = formated_msg.split(":::")
        if self.exclude.logger and not logger_name.startswith(self.exclude.logger):
            return False

        match = self.regex.match(log_msg)
        if not match:
            _logger.debug("Regex filter out")
            # Line don't match regex, return False to propagate to other filter
            return False
        logging.getLogger("odoo.runbot.filter").info(
            "%s matched %s for logger %s",
            self.exclude.name,
            bool(match),
            logger_name,
        )
        self.log_match.append(original_record)
        return True

    @property
    def success(self) -> bool:
        return self.exclude.min_match <= len(self.log_match) <= self.exclude.max_match

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(exclude={self.exclude!r})"

    def reset_counter(self) -> None:
        self.log_match = []


@contextlib.contextmanager
def start_warning_log_watcher(step: RunbotStepConfig) -> Callable[[], list[ExcludeWarningFilter]]:
    _logger.debug("Starting odoo logging interceptor with %s regex", len(step.log_filters))
    runbot_handler = get_handler() or RunbotWarningWatcherHandler()  # Auto registering
    runbot_handler.add_warnings(step.log_filters)
    logging.getLogger().addHandler(runbot_handler)
    yield partial(get_warning_log_watcher, step)


def get_warning_log_watcher(step: RunbotStepConfig = None) -> list[ExcludeWarningFilter]:
    runbot_handler = get_handler()
    if not runbot_handler:
        return []
    if not step:
        return runbot_handler.remove_warnings()
    return runbot_handler.remove_warnings(step.log_filters)



import logging
import warnings

from odoo import models

_DEFAULT_LOGGER_NAME = __name__

logging.getLogger(_DEFAULT_LOGGER_NAME).warning("Runbot Test Warmup Warning")


class SimpleLogger(models.AbstractModel):
    _name = "runbot.logger"
    _description = "Logging Tester"

    def logger(self, logger_name=_DEFAULT_LOGGER_NAME):
        return logging.getLogger(logger_name)

    def do_pywarnings(self, msg, warning_type):
        warnings.warn(msg, warning_type, stacklevel=2)



"""
The *pathspec.patterns* package contains the pattern matching
implementations.
"""

# Load pattern implementations.
from . import gitwildmatch

# DEPRECATED: Expose the `GitWildMatchPattern` class in this module for
# backward compatibility with v0.5.
from .gitwildmatch import GitWildMatchPattern



from __future__ import annotations

import builtins
import logging
import typing
import warnings

import coverage
from odoo_env_config import api, entry

from . import _odoo_internal, runbot_run_logging, runbot_run_test
from .runbot_env import RunbotPyWarningsFilter, StepAction

if typing.TYPE_CHECKING:
    from pathlib import Path
    from warnings import WarningMessage

    from .runbot_env import RunbotEnvironment, RunbotStepConfig

_logger = logging.getLogger("odoo_runbot")

__all__ = ["StepRunner"]


class StepRunner:
    """The main logique to execiute a runbot step-by-step. See `run`"""

    def __init__(self, env: RunbotEnvironment) -> None:
        self.env = env
        self.has_run: bool = False

    def get_server_args_from_step(self, step: RunbotStepConfig) -> list[str]:
        update_install = entry.env_to_section(entry.UpdateInstallSection)
        misc = entry.env_to_section(entry.MiscSection)
        http_sec = entry.env_to_section(entry.HttpOdooConfigSection)
        test_sec = entry.env_to_section(entry.TestOdooConfigSection)
        update_install.install = step.modules
        misc.stop_after_init = True
        http_sec.interface = "127.0.0.1"
        http_sec.enable = True
        test_sec.enable = step.action == StepAction.TESTS
        if test_sec.enable and step.test_tags:
            test_sec.test_tags = ",".join(step.test_tags)
        values_update = {}
        values_update.update(test_sec.to_values())
        values_update.update(update_install.to_values())
        values_update.update(misc.to_values())
        values_update.update(http_sec.to_values())
        return api.dict_to_odoo_args(values_update)

    def setup_odoo(self) -> None:
        _odoo_internal.setup_odoo()

    def execute(self, step: RunbotStepConfig) -> int:
        """The module to install before runnning the test.
        This modules can be a `CSV` value to install multiple value.
        Usage:
            module_name="account,my_project_config"
        Warning: When this variable is used, the test are disabled, so all the test with tag `+post_install` are not run
        """
        _logger.info("Run step %s(action=%s)", step.name, str(step.action.value))
        self.has_run = True
        if step.action == StepAction.INSTALL:
            return self._do_execute_action_install(step)
        if step.action == StepAction.TESTS:
            return self._do_execute_action_tests(step)
        _logger.error("Unknown action: %s", step.action)
        return 404

    def _do_execute_action_install(self, step: RunbotStepConfig) -> int:
        """The module to install before runnning the test.
        This modules can be a `CSV` value to install multiple value.
        Usage:
            module_name="account,my_project_config"
        Warning: When this variable is used, the test are disabled, so all the test with tag `+post_install` are not run
        """
        _odoo_internal.parse_config(self.get_server_args_from_step(step))
        output_dir = self.env.result_path / "x_unittest"
        with runbot_run_logging.start_warning_log_watcher(step) as log_filters_extractor:  # noqa: SIM117
            with warnings.catch_warnings(record=True) as warnings_list:
                rc = _odoo_internal.run_odoo_and_stop()
                _logger.info("Odoo Done rc=%s", rc)
        log_filters = log_filters_extractor()
        return abs(rc) + self._execute_to_filters(step, log_filters, warnings_list, output_dir=output_dir)

    def _do_execute_action_tests(self, step: RunbotStepConfig) -> int:
        """Run a step and activate all the test feat:
        - coverage
        - Logger matching
        - Xunit report with [xmlrunner](https://unittest-xml-reporting.readthedocs.io/en/latest/)

        Args:
            step: The step to run, containing log filter and warning filter

        Returns: The result code > 1 mean error

        """
        coverage_watcher = self.get_coverage()
        _logger.info("Setup Odoo ...")
        _odoo_internal.parse_config(self.get_server_args_from_step(step))

        logging.getLogger("odoo.tests.stats").setLevel(logging.DEBUG)
        output_dir = self.env.result_path / "x_unittest"
        _logger.info("Start Coverage recording...")
        with runbot_run_logging.start_warning_log_watcher(step) as log_filters_extractor:  # noqa: SIM117
            with _odoo_internal.patch_odoo_test_suite(output_dir):
                with warnings.catch_warnings(record=True) as warnings_list:
                    with coverage_watcher.collect():
                        rc = _odoo_internal.run_odoo_and_stop()
                        _logger.info("Odoo Done rc=%s", rc)
        log_filters = log_filters_extractor()
        _logger.info("Save Coverage recording in [%s] ...", coverage_watcher.config.data_file)
        if not abs(rc):
            coverage_watcher.report()
        return abs(rc) + self._execute_to_filters(step, log_filters, warnings_list, output_dir=output_dir)

    def get_coverage(self) -> coverage.Coverage:
        debug = None
        if self.env.verbose:
            debug = ["config"]
        return coverage.Coverage(
            source=[str(self.env.abs_curr_dir)],
            branch=True,
            debug=debug,
            auto_data=True,
            config_file=str(self.env.abs_curr_dir.joinpath("pyproject.toml")),
            omit=["**/__manifest__.py", "**/__init__.py"],
            data_file=self.env.result_path / ".coverage",
        )

    def _execute_to_filters(
        self,
        step: RunbotStepConfig,
        log_filters: list[runbot_run_logging.ExcludeWarningFilter],
        warnings_list: list[WarningMessage],
        output_dir: Path,
    ) -> int:
        if log_filters or warnings_list:
            _logger.info("Test Logger and py.warnings...")
            res = runbot_run_test.execute_test_after_odoo(
                step.name,
                log_filters,
                warnings_list if not step.allow_warnings else [],
                test_runner=runbot_run_test.get_test_runner(output_dir),
            )
            rc = int(not res.wasSuccessful())
            _logger.info("Test logger rc=%s", rc)
            return rc
        return 0

    def setup_warning_filter(self, warning_filters: list[RunbotPyWarningsFilter]) -> None:
        for waring_filer in warning_filters:
            warnings.filterwarnings(
                action=waring_filer.action,
                category=getattr(builtins, waring_filer.category, Warning),
                message=waring_filer.message,
            )



"""
This module provides an object oriented interface for pattern matching of files.
"""

from collections.abc import (
	Collection as CollectionType)
from itertools import (
	zip_longest)
from typing import (
	AnyStr,
	Callable,  # Replaced by `collections.abc.Callable` in 3.9.
	Collection,  # Replaced by `collections.abc.Collection` in 3.9.
	Iterable,  # Replaced by `collections.abc.Iterable` in 3.9.
	Iterator,  # Replaced by `collections.abc.Iterator` in 3.9.
	Optional,  # Replaced by `X | None` in 3.10.
	Type,  # Replaced by `type` in 3.9.
	TypeVar,
	Union)  # Replaced by `X | Y` in 3.10.

from . import util
from .pattern import (
	Pattern)
from .util import (
	CheckResult,
	StrPath,
	TStrPath,
	TreeEntry,
	_filter_check_patterns,
	_is_iterable,
	normalize_file)

Self = TypeVar("Self", bound="PathSpec")
"""
:class:`PathSpec` self type hint to support Python v<3.11 using PEP 673
recommendation.
"""


class PathSpec(object):
	"""
	The :class:`PathSpec` class is a wrapper around a list of compiled
	:class:`.Pattern` instances.
	"""

	def __init__(self, patterns: Iterable[Pattern]) -> None:
		"""
		Initializes the :class:`PathSpec` instance.

		*patterns* (:class:`~collections.abc.Collection` or :class:`~collections.abc.Iterable`)
		yields each compiled pattern (:class:`.Pattern`).
		"""
		if not isinstance(patterns, CollectionType):
			patterns = list(patterns)

		self.patterns: Collection[Pattern] = patterns
		"""
		*patterns* (:class:`~collections.abc.Collection` of :class:`.Pattern`)
		contains the compiled patterns.
		"""

	def __eq__(self, other: object) -> bool:
		"""
		Tests the equality of this path-spec with *other* (:class:`PathSpec`)
		by comparing their :attr:`~PathSpec.patterns` attributes.
		"""
		if isinstance(other, PathSpec):
			paired_patterns = zip_longest(self.patterns, other.patterns)
			return all(a == b for a, b in paired_patterns)
		else:
			return NotImplemented

	def __len__(self) -> int:
		"""
		Returns the number of compiled patterns this path-spec contains
		(:class:`int`).
		"""
		return len(self.patterns)

	def __add__(self: Self, other: "PathSpec") -> Self:
		"""
		Combines the :attr:`Pathspec.patterns` patterns from two
		:class:`PathSpec` instances.
		"""
		if isinstance(other, PathSpec):
			return self.__class__(self.patterns + other.patterns)
		else:
			return NotImplemented

	def __iadd__(self: Self, other: "PathSpec") -> Self:
		"""
		Adds the :attr:`Pathspec.patterns` patterns from one :class:`PathSpec`
		instance to this instance.
		"""
		if isinstance(other, PathSpec):
			self.patterns += other.patterns
			return self
		else:
			return NotImplemented

	def check_file(
		self,
		file: TStrPath,
		separators: Optional[Collection[str]] = None,
	) -> CheckResult[TStrPath]:
		"""
		Check the files against this path-spec.

		*file* (:class:`str` or :class:`os.PathLike`) is the file path to be
		matched against :attr:`self.patterns <PathSpec.patterns>`.

		*separators* (:class:`~collections.abc.Collection` of :class:`str`; or
		:data:`None`) optionally contains the path separators to normalize. See
		:func:`~pathspec.util.normalize_file` for more information.

		Returns the file check result (:class:`~pathspec.util.CheckResult`).
		"""
		norm_file = normalize_file(file, separators)
		include, index = self._match_file(enumerate(self.patterns), norm_file)
		return CheckResult(file, include, index)

	def check_files(
		self,
		files: Iterable[TStrPath],
		separators: Optional[Collection[str]] = None,
	) -> Iterator[CheckResult[TStrPath]]:
		"""
		Check the files against this path-spec.

		*files* (:class:`~collections.abc.Iterable` of :class:`str` or
		:class:`os.PathLike`) contains the file paths to be checked against
		:attr:`self.patterns <PathSpec.patterns>`.

		*separators* (:class:`~collections.abc.Collection` of :class:`str`; or
		:data:`None`) optionally contains the path separators to normalize. See
		:func:`~pathspec.util.normalize_file` for more information.

		Returns an :class:`~collections.abc.Iterator` yielding each file check
		result (:class:`~pathspec.util.CheckResult`).
		"""
		if not _is_iterable(files):
			raise TypeError(f"files:{files!r} is not an iterable.")

		use_patterns = _filter_check_patterns(self.patterns)
		for orig_file in files:
			norm_file = normalize_file(orig_file, separators)
			include, index = self._match_file(use_patterns, norm_file)
			yield CheckResult(orig_file, include, index)

	def check_tree_files(
		self,
		root: StrPath,
		on_error: Optional[Callable[[OSError], None]] = None,
		follow_links: Optional[bool] = None,
	) -> Iterator[CheckResult[str]]:
		"""
		Walks the specified root path for all files and checks them against this
		path-spec.

		*root* (:class:`str` or :class:`os.PathLike`) is the root directory to
		search for files.

		*on_error* (:class:`~collections.abc.Callable` or :data:`None`) optionally
		is the error handler for file-system exceptions. It will be called with the
		exception (:exc:`OSError`). Reraise the exception to abort the walk. Default
		is :data:`None` to ignore file-system exceptions.

		*follow_links* (:class:`bool` or :data:`None`) optionally is whether to walk
		symbolic links that resolve to directories. Default is :data:`None` for
		:data:`True`.

		*negate* (:class:`bool` or :data:`None`) is whether to negate the match
		results of the patterns. If :data:`True`, a pattern matching a file will
		exclude the file rather than include it. Default is :data:`None` for
		:data:`False`.

		Returns an :class:`~collections.abc.Iterator` yielding each file check
		result (:class:`~pathspec.util.CheckResult`).
		"""
		files = util.iter_tree_files(root, on_error=on_error, follow_links=follow_links)
		yield from self.check_files(files)

	@classmethod
	def from_lines(
		cls: Type[Self],
		pattern_factory: Union[str, Callable[[AnyStr], Pattern]],
		lines: Iterable[AnyStr],
	) -> Self:
		"""
		Compiles the pattern lines.

		*pattern_factory* can be either the name of a registered pattern factory
		(:class:`str`), or a :class:`~collections.abc.Callable` used to compile
		patterns. It must accept an uncompiled pattern (:class:`str`) and return the
		compiled pattern (:class:`.Pattern`).

		*lines* (:class:`~collections.abc.Iterable`) yields each uncompiled pattern
		(:class:`str`). This simply has to yield each line so that it can be a
		:class:`io.TextIOBase` (e.g., from :func:`open` or :class:`io.StringIO`) or
		the result from :meth:`str.splitlines`.

		Returns the :class:`PathSpec` instance.
		"""
		if isinstance(pattern_factory, str):
			pattern_factory = util.lookup_pattern(pattern_factory)

		if not callable(pattern_factory):
			raise TypeError(f"pattern_factory:{pattern_factory!r} is not callable.")

		if not _is_iterable(lines):
			raise TypeError(f"lines:{lines!r} is not an iterable.")

		patterns = [pattern_factory(line) for line in lines if line]
		return cls(patterns)

	def match_entries(
		self,
		entries: Iterable[TreeEntry],
		separators: Optional[Collection[str]] = None,
		*,
		negate: Optional[bool] = None,
	) -> Iterator[TreeEntry]:
		"""
		Matches the entries to this path-spec.

		*entries* (:class:`~collections.abc.Iterable` of :class:`~pathspec.util.TreeEntry`)
		contains the entries to be matched against :attr:`self.patterns <PathSpec.patterns>`.

		*separators* (:class:`~collections.abc.Collection` of :class:`str`; or
		:data:`None`) optionally contains the path separators to normalize. See
		:func:`~pathspec.util.normalize_file` for more information.

		*negate* (:class:`bool` or :data:`None`) is whether to negate the match
		results of the patterns. If :data:`True`, a pattern matching a file will
		exclude the file rather than include it. Default is :data:`None` for
		:data:`False`.

		Returns the matched entries (:class:`~collections.abc.Iterator` of
		:class:`~pathspec.util.TreeEntry`).
		"""
		if not _is_iterable(entries):
			raise TypeError(f"entries:{entries!r} is not an iterable.")

		use_patterns = _filter_check_patterns(self.patterns)
		for entry in entries:
			norm_file = normalize_file(entry.path, separators)
			include, _index = self._match_file(use_patterns, norm_file)

			if negate:
				include = not include

			if include:
				yield entry

	_match_file = staticmethod(util.check_match_file)
	"""
	Match files using the `check_match_file()` utility function. Subclasses may
	override this method as an instance method. It does not have to be a static
	method. The signature for this method is subject to change.
	"""

	def match_file(
		self,
		file: StrPath,
		separators: Optional[Collection[str]] = None,
	) -> bool:
		"""
		Matches the file to this path-spec.

		*file* (:class:`str` or :class:`os.PathLike`) is the file path to be
		matched against :attr:`self.patterns <PathSpec.patterns>`.

		*separators* (:class:`~collections.abc.Collection` of :class:`str`)
		optionally contains the path separators to normalize. See
		:func:`~pathspec.util.normalize_file` for more information.

		Returns :data:`True` if *file* matched; otherwise, :data:`False`.
		"""
		norm_file = normalize_file(file, separators)
		include, _index = self._match_file(enumerate(self.patterns), norm_file)
		return bool(include)

	def match_files(
		self,
		files: Iterable[StrPath],
		separators: Optional[Collection[str]] = None,
		*,
		negate: Optional[bool] = None,
	) -> Iterator[StrPath]:
		"""
		Matches the files to this path-spec.

		*files* (:class:`~collections.abc.Iterable` of :class:`str` or
		:class:`os.PathLike`) contains the file paths to be matched against
		:attr:`self.patterns <PathSpec.patterns>`.

		*separators* (:class:`~collections.abc.Collection` of :class:`str`; or
		:data:`None`) optionally contains the path separators to normalize. See
		:func:`~pathspec.util.normalize_file` for more information.

		*negate* (:class:`bool` or :data:`None`) is whether to negate the match
		results of the patterns. If :data:`True`, a pattern matching a file will
		exclude the file rather than include it. Default is :data:`None` for
		:data:`False`.

		Returns the matched files (:class:`~collections.abc.Iterator` of
		:class:`str` or :class:`os.PathLike`).
		"""
		if not _is_iterable(files):
			raise TypeError(f"files:{files!r} is not an iterable.")

		use_patterns = _filter_check_patterns(self.patterns)
		for orig_file in files:
			norm_file = normalize_file(orig_file, separators)
			include, _index = self._match_file(use_patterns, norm_file)

			if negate:
				include = not include

			if include:
				yield orig_file

	def match_tree_entries(
		self,
		root: StrPath,
		on_error: Optional[Callable[[OSError], None]] = None,
		follow_links: Optional[bool] = None,
		*,
		negate: Optional[bool] = None,
	) -> Iterator[TreeEntry]:
		"""
		Walks the specified root path for all files and matches them to this
		path-spec.

		*root* (:class:`str` or :class:`os.PathLike`) is the root directory to
		search.

		*on_error* (:class:`~collections.abc.Callable` or :data:`None`) optionally
		is the error handler for file-system exceptions. It will be called with the
		exception (:exc:`OSError`). Reraise the exception to abort the walk. Default
		is :data:`None` to ignore file-system exceptions.

		*follow_links* (:class:`bool` or :data:`None`) optionally is whether to walk
		symbolic links that resolve to directories. Default is :data:`None` for
		:data:`True`.

		*negate* (:class:`bool` or :data:`None`) is whether to negate the match
		results of the patterns. If :data:`True`, a pattern matching a file will
		exclude the file rather than include it. Default is :data:`None` for
		:data:`False`.

		Returns the matched files (:class:`~collections.abc.Iterator` of
		:class:`.TreeEntry`).
		"""
		entries = util.iter_tree_entries(root, on_error=on_error, follow_links=follow_links)
		yield from self.match_entries(entries, negate=negate)

	def match_tree_files(
		self,
		root: StrPath,
		on_error: Optional[Callable[[OSError], None]] = None,
		follow_links: Optional[bool] = None,
		*,
		negate: Optional[bool] = None,
	) -> Iterator[str]:
		"""
		Walks the specified root path for all files and matches them to this
		path-spec.

		*root* (:class:`str` or :class:`os.PathLike`) is the root directory to
		search for files.

		*on_error* (:class:`~collections.abc.Callable` or :data:`None`) optionally
		is the error handler for file-system exceptions. It will be called with the
		exception (:exc:`OSError`). Reraise the exception to abort the walk. Default
		is :data:`None` to ignore file-system exceptions.

		*follow_links* (:class:`bool` or :data:`None`) optionally is whether to walk
		symbolic links that resolve to directories. Default is :data:`None` for
		:data:`True`.

		*negate* (:class:`bool` or :data:`None`) is whether to negate the match
		results of the patterns. If :data:`True`, a pattern matching a file will
		exclude the file rather than include it. Default is :data:`None` for
		:data:`False`.

		Returns the matched files (:class:`~collections.abc.Iterable` of
		:class:`str`).
		"""
		files = util.iter_tree_files(root, on_error=on_error, follow_links=follow_links)
		yield from self.match_files(files, negate=negate)

	# Alias `match_tree_files()` as `match_tree()` for backward compatibility
	# before v0.3.2.
	match_tree = match_tree_files



from __future__ import annotations

from typing import Any

from hatchling.builders.binary import BinaryBuilder


class AppBuilder(BinaryBuilder):
    PLUGIN_NAME = 'app'

    def build_bootstrap(
        self,
        directory: str,
        **build_data: Any,
    ) -> str:
        self.app.display_warning(
            'The `app` build target is deprecated and will be removed in a future release. '
            'Use the `binary` build target instead.'
        )
        return super().build_bootstrap(directory, **build_data)



from __future__ import annotations

import contextlib
import logging
import typing
import unittest

if typing.TYPE_CHECKING:
    import pathlib

from unittest.mock import patch

from xmlrunner import XMLTestRunner
from xmlrunner.result import _TestInfo, _XMLTestResult

_logger = logging.getLogger("odoo_runbot.odoo_internal")
try:
    import odoo

except ModuleNotFoundError:
    from unittest import mock

    odoo = mock.MagicMock()

__all__ = ["odoo", "patch_odoo_test_suite", "run_odoo_and_stop", "setup_odoo"]


class _RunProto(typing.Protocol):
    def run(self, result): ...


class _UpdateProto(typing.Protocol):
    def update(self, other): ...


class _XmlTestInfo(_TestInfo):
    def shortDescription(self):  # noqa: N802
        """
        Taken from unittest.TestCase.shortDescription
        Returns a one-line description of the test, or None if no
        description has been provided.

        The default implementation of this method returns the first line of
        the specified test method's docstring.
        """
        return self.doc.strip().split("\n")[0].strip() if self.doc else None


class _XmlTestResult(_XMLTestResult):
    def __init__(self, *args: list, **kwargs: dict) -> None:
        super().__init__(*args, **kwargs, infoclass=_XmlTestInfo)


@contextlib.contextmanager
def patch_odoo_test_suite(output_dir: pathlib.Path) -> typing.Generator[None]:
    """Patch odoo test suite and Odoo test result.

    - In Odoo 12 and Odoo 13, The test result is the native unittest.TestResult
    - In Odoo 14 and Odoo 15. The test result extend unittest.TestResult and add `update` function. So we call it
    - In 16.0 and more the OdooTestResult don't extend unittest.TestResult.
    I need to patch the `update` function to add result of the `_XmlTestResult`

    Args:
        output_dir: Te patch where to store the xml result

    Returns: Nothing, but unpatched function at the end
    """

    odoo_suite_class, odoo_result_class = _try_find_stuff_to_patch()
    if not odoo_suite_class:
        msg = f"Odoo version {odoo.release.version} not supported"
        raise ValueError(msg)
    output_dir.mkdir(exist_ok=True, parents=True)

    _logger.info("Create test result in %s", output_dir.absolute())
    _unpatched_run = odoo_suite_class.run
    patch_odoo_suite_run = f"{odoo_suite_class.__module__}.{odoo_suite_class.__name__}.run"
    _logger.info("Patch %s", patch_odoo_suite_run)
    patch_odoo_suite_update = None
    if odoo_result_class:
        _unpatched_update = odoo_result_class.update
        patch_odoo_suite_update = f"{odoo_result_class.__module__}.{odoo_result_class.__name__}.update"
        _logger.info("Patch %s", patch_odoo_suite_update)

    def patched_suite_run(self, result):
        # Suite run method will be called by the XMLTestRunner,
        # so we need to run the original run method
        with patch.object(self, "run", lambda result: _unpatched_run(self, result)):
            # Override : XMLTestRunner to run the tests and generate XML reports
            xml_results = XMLTestRunner(output=str(output_dir.absolute()), verbosity=2, resultclass=_XmlTestResult).run(
                self
            )
            _logger.info(xml_results)
        if hasattr(result, "update"):
            result.update(xml_results)
        elif isinstance(result, unittest.TestResult):
            result.failfast = xml_results.failfast
            result.failures = xml_results.failures
            result.testsRun = xml_results.testsRun
            result.errors = xml_results.errors
            result.skipped = xml_results.skipped
        return result

    def patched_suite_update(self, other):
        # Adapt _XMLTestResult to OdooTestResult
        if isinstance(other, _XMLTestResult):
            self.failures_count += len(other.failures)
            self.errors_count += len(other.errors)
            self.skipped += len(other.skipped)
            self.testsRun += other.testsRun
        else:
            _unpatched_update(self, other)

    with patch(patch_odoo_suite_run, patched_suite_run):
        if odoo_result_class:
            with patch(patch_odoo_suite_update, patched_suite_update):
                yield
        else:
            yield


def _try_find_stuff_to_patch() -> tuple[type[_RunProto] | None, type[_UpdateProto] | None]:
    if "12.0" <= odoo.release.version <= "13.0" or "14.0" <= odoo.release.version <= "15.0":
        from odoo.tests.common import OdooSuite as odoo_suite_class  # noqa: N813

        odoo_result_class = None
    elif "16.0" <= odoo.release.version <= "18.0":
        from odoo.tests.result import OdooTestResult as odoo_result_class  # noqa: N813
        from odoo.tests.suite import OdooSuite as odoo_suite_class  # noqa: N813
    else:
        return None, None

    return odoo_suite_class, odoo_result_class


def run_odoo_and_stop(preload: list[str] | None = None) -> int:
    preload = preload or odoo.tools.config["db_name"].split(",")
    return odoo.service.server.start(preload=preload, stop=True)


def setup_odoo() -> None:
    odoo.cli.server.check_root_user()
    odoo.cli.server.check_postgres_user()
    odoo.cli.server.report_configuration()
    odoo.netsvc.init_logger()


def parse_config(server_args: list[str]) -> None:
    __import__("os").environ["TZ"] = "UTC"
    _logger.info("Start Odoo with args=%s", server_args)
    # Hard set value
    # In test mode we don't care about loading language
    # But il you do multiple step, you load multiple time the registry.
    # And in odoo.modules.loadings.load_module they `pop` this key, idkw
    # So in test mode, we force this option key to None
    odoo.tools.config.options["load_language"] = None
    odoo.tools.config.parse_config(server_args)



from __future__ import annotations

import configparser
import importlib
import logging
import time
import typing
from pathlib import Path

import odoo_env_config
from addons_installer.addons_installer import AddonsFinder, AddonsInstaller, OdooAddonsDef
from odoo_env_config import entry, section

if typing.TYPE_CHECKING:
    from rich.console import ConsoleRenderable

    from .runbot_env import RunbotEnvironment

from rich.table import Table

from . import _odoo_internal
from .runbot_init_branch_resolver import AddonsBranchResolver, AddonsToInstall

_logger = logging.getLogger("odoo_runbot")


def show_addons(env: RunbotEnvironment) -> ConsoleRenderable:
    """Returns:"""
    _logger.debug("Find debpends addons from environement variables")
    addons: set[OdooAddonsDef] = AddonsFinder().parse_env(env.environ)
    addons_branch_resolver = AddonsBranchResolver(env, addons)
    addons_to_install: list[AddonsToInstall] = addons_branch_resolver.resolve_branch_addons_git()
    table = Table("Name", "Branch to test", "Used", "Install Path", "Exist")
    for to_install in addons_to_install:
        exist = ":heavy_check_mark:" if Path(to_install.addon.addons_path).exists() else ":x:"
        table.add_row(
            to_install.addon.name,
            ", ".join(to_install.branch_try),
            to_install.branch_to_use,
            to_install.addon.addons_path,
            exist,
        )
    return table, addons_to_install


def install_addons(addons_to_install: list[AddonsToInstall]) -> None:
    for addon in addons_to_install:
        AddonsInstaller.install(addon.addon)


def init_database(env: RunbotEnvironment, max_retry: int = 3) -> int:
    """Returns:"""
    if not importlib.util.find_spec("psycopg2"):
        _logger.warning("psycopg2 not installed")

    db_section = entry.env_to_section(section.DatabaseOdooConfigSection, extra_env=env.environ)
    _logger.info("Creating database %s on %s:%s...", db_section.name, db_section.host, db_section.port)
    if _db_exist(db_section):
        _logger.info("Database %s on %s:%s already created", db_section.name, db_section.host, db_section.port)
        return 0

    import psycopg2
    from psycopg2 import sql

    conn = psycopg2.connect(
        dbname=db_section.user,
        user=db_section.user,
        password=db_section.password,
        host=db_section.host,
        port=db_section.port,
    )
    conn.autocommit = True
    retry = 0
    db_created = False
    createdb_cmd = sql.SQL("CREATE DATABASE {} ENCODING 'unicode' LC_COLLATE 'C' TEMPLATE template0;").format(
        sql.Identifier(db_section.name),
    )
    while retry <= max_retry and not db_created:
        time.sleep(5)
        try:
            with conn.cursor() as curs:
                curs.execute(createdb_cmd)
                db_created = True
                _logger.info(
                    "Database %s on %s:%s created successfully",
                    db_section.name,
                    db_section.host,
                    db_section.port,
                )
        except psycopg2.OperationalError:
            retry += 1
            _logger.info(
                "Failed to create database %s on %s:%s. retry in 5 seconds (%s / %s)",
                db_section.name,
                db_section.host,
                db_section.port,
                retry,
                max_retry,
            )
            if retry > max_retry:
                raise
    return 0


def init_config(env: RunbotEnvironment) -> ConsoleRenderable:
    """Returns:"""
    odoo_env_config.entry.direct_run_command(_odoo_internal.odoo, [], env.ODOO_RC, other_env=env.environ)
    config = configparser.RawConfigParser()
    config.read(env.ODOO_RC)

    table = Table("property", "value", title="options")
    for ini_section in config.sections():
        for option in config.options(ini_section):
            table.add_row(option, config.get(ini_section, option))
    _odoo_internal.odoo.netsvc.init_logger()
    logging.getLogger("odoo").setLevel(logging.INFO)
    _logger.info("Init Logging for Odoo")
    return table


def _db_exist(db_section: section.DatabaseOdooConfigSection) -> bool:
    """Args:
        db_section:

    Returns:

    """
    import psycopg2
    from psycopg2 import sql

    try:
        conn = psycopg2.connect(
            dbname=db_section.name,
            user=db_section.user,
            password=db_section.password,
            host=db_section.host,
            port=db_section.port,
        )
        with conn, conn.cursor() as curs:
            curs.execute(sql.SQL("SELECT 1;"))
            return True
    except psycopg2.OperationalError:
        return False



import logging

import odoo
from odoo.tests import common
from odoo.tools.misc import mute_logger


class TestStep2(common.TransactionCase):
    """
    Test log capturing and py.warnings capturing in unitest don't need to be captured by step filter
    """

    # In Odoo prior to 15, the complete name of the logger is required
    @mute_logger("odoo.addons.runbot_tester.models.runbot_tester")
    def test_logger1(self):
        """
        Assert `odoo.tools.misc.mute_logger` catch all log line
        If not, the step should failed, because no step filter are avaliable
        """
        self.env["runbot.logger"].logger().info("My info message")
        self.env["runbot.logger"].logger().warning("My warning message")
        self.env["runbot.logger"].logger().critical("My critical message")
        self.env["runbot.logger"].logger().error("My error message")
        self.assertTrue(True, "Dummy assert to be sure this test is passed")

    def test_logger2(self):
        """
        Test when self.assertLogs() is used, you don't need to add a regex on the runbot
        We test with root logger, custom non exitant logger, and classic logger
        Test [unitest.assertLogs][] work as expected
        """
        with self.assertLogs() as logs:
            self.env["runbot.logger"].logger(logger_name=None).warning("My %s", "message1")
        self.assertEqual(logs.output, ["WARNING:root:My message1"])

        with self.assertLogs("odoo.addons.runbot_tester.models.runbot_tester") as l2:
            self.env["runbot.logger"].logger().error("My message2")
        self.assertEqual(l2.output, ["ERROR:odoo.addons.runbot_tester.models.runbot_tester:My message2"])

        with self.assertLogs("custom.logger") as l3:
            self.env["runbot.logger"].logger("custom.logger").critical("My message3")
        self.assertEqual(l3.output, ["CRITICAL:custom.logger:My message3"])

    def test_logger3(self):
        """
        Log message with allowed level in runbot
        - INFO
        - DEBUG
        - RUNBOT (level added by Odoo in test mode)
        """
        self.env["runbot.logger"].logger().info("message INFO")
        self.env["runbot.logger"].logger().debug("message DEBUG")
        if odoo.release.version >= "14.0":
            # Logger level RUNBOT activated in odoo 14 and more
            self.env["runbot.logger"].logger().log(logging.RUNBOT, "message RUNBOT")

    def test_logger4(self):
        """
        Assert you don't need to add a warning-filter on your runbot.step if you do the filtering in your test.
        Test [unitest.assertWarnsRegex][] and [unittest.assertWarns][] works as expected
        """
        with self.assertWarnsRegex(DeprecationWarning, expected_regex="This is depreca.*"):
            self.env["runbot.logger"].do_pywarnings("This is deprecated", DeprecationWarning)
        with self.assertWarns(BytesWarning):
            self.env["runbot.logger"].do_pywarnings("This is deprecated", BytesWarning)



from odoo.tests import common


class TestStep1(common.TransactionCase):
    """
    A simple to ensure odoo demo data are loaded
    Odoo can't run test without demo, but we have a warmup step (install this module without test-enable)
    So we ensure that odoo demo data are loaded even without test-enable
    """

    def test_step1(self):
        self.assertTrue(self.env.ref("base.user_demo"), "Big trouble, no demo data in your database ?")



"""This lib is a CLI using typer.

I tried to separate the code by feature part.

- [odoo_runbot.cli][] Contains the entrypoint, All the rich display is done here. (Other than log)
- [odoo_runbot.runbot_env][] Contains il a python way the env in wich the runbot is running.
- [odoo_runbot.runbot_init][] Contains all the code to execute to initialize the Odoo to run test
- [odoo_runbot.runbot_init_branch_resolver][] Contains the working branch resolving (Pull other add-ons from git)
"""

__import__("os").environ["TZ"] = "UTC"
__import__("os").environ["WITHOUT_DEMO"] = "False"



import pathlib

from odoo_runbot import runbot_env
from odoo_runbot.runbot_env import RunbotExcludeWarning, RunbotStepConfig, RunbotToolConfig, StepAction


def sample_config(fname: str) -> pathlib.Path:
    return pathlib.Path(__file__).resolve().parent.joinpath("sample_config", fname)


def test_minimal_config():
    """[tool.runbot]
    modules = ["module_to_test"]
    """
    config_path = sample_config("pyproject_minimal.toml")
    config = runbot_env.RunbotToolConfig.load_from_toml(config_path)
    global_module = ["module_to_test"]
    assert config == RunbotToolConfig(
        include_current_project=True,
        steps=[
            RunbotStepConfig(
                name="default",
                modules=global_module,
                action=StepAction.TESTS,
                test_tags=[],
                coverage=True,
                log_filters=[],
            ),
        ],
        pretty=True,
    )


def test_pyproject_classic():
    global_module = ["module_to_test", "module_to_test_2"]
    global_log_filter = [
        RunbotExcludeWarning(
            name="All Steps - Logger Filter 1",
            regex=".*log to accept.*",
            logger="",
            min_match=1,
            max_match=1,
        )
    ]

    config = runbot_env.RunbotToolConfig.load_from_toml(sample_config("pyproject_classic.toml"))
    config_full = runbot_env.RunbotToolConfig.load_from_toml(sample_config("pyproject_classic.full.toml"))
    assert (
        config
        == config_full
        == RunbotToolConfig(
            include_current_project=True,
            steps=[
                RunbotStepConfig(
                    name="install",
                    modules=global_module,
                    action=StepAction.INSTALL,
                    test_tags=[],
                    coverage=False,
                    log_filters=global_log_filter,
                ),
                RunbotStepConfig(
                    name="run-test",
                    modules=global_module,
                    action=StepAction.TESTS,
                    test_tags=["/module_to_test:MyTestCase", "/module_to_test"],
                    coverage=True,
                    log_filters=global_log_filter,
                ),
            ],
            pretty=True,
        )
    )


def test_pyproject_complex():
    config = runbot_env.RunbotToolConfig.load_from_toml(sample_config("pyproject_complex.toml"))

    global_regex = [
        runbot_env.RunbotExcludeWarning(
            name="All Steps - Logger Filter 1",
            regex=r".*global-regex-warning-1.*",
        ),
        runbot_env.RunbotExcludeWarning(
            name="global-regex-warning-2",
            regex=r".*global-regex-warning-2.*",
            min_match=1,
            max_match=2,
        ),
    ]
    global_module = ["module_to_test", "module_to_test2"]
    global_coverage = False

    assert config == RunbotToolConfig(
        include_current_project=True,
        steps=[
            RunbotStepConfig(
                name="install",
                modules=["first_module_to_install"],
                action=StepAction.INSTALL,
                test_tags=[],
                coverage=global_coverage,
                log_filters=[
                    *global_regex,
                    RunbotExcludeWarning(
                        regex=".*Install filter.*",
                        name="Step install - Logger Filter 3",
                        min_match=1,
                        max_match=1,
                    ),
                ],
            ),
            RunbotStepConfig(
                name="tests",
                modules=global_module,
                action=StepAction.TESTS,
                test_tags=["+at-install", "-post-install"],
                coverage=True,
                log_filters=[
                    *global_regex,
                    RunbotExcludeWarning(
                        regex=".*regex warning.*",
                        name="test-regex-log-warning-2",
                        min_match=1,
                        max_match=1,
                    ),
                ],
            ),
            RunbotStepConfig(
                name="warmup",
                modules=["second_module_to_install"],
                action=StepAction.INSTALL,
                test_tags=[],
                coverage=global_coverage,
                log_filters=global_regex,
            ),
            RunbotStepConfig(
                name="Post install test",
                modules=["module_to_test2"],
                action=StepAction.TESTS,
                test_tags=["-at-install", "+post-install"],
                coverage=False,
                log_filters=[
                    *global_regex,
                    RunbotExcludeWarning(
                        name="Step Post install test - Logger Filter 3",
                        regex=".*Post install test regex-warnings.*",
                        min_match=2,
                        max_match=2,
                    ),
                ],
            ),
        ],
        pretty=True,
    )


def test_min_max_match_log_filter():
    assert RunbotExcludeWarning(name="A", regex="A", min_match=2) == RunbotExcludeWarning(
        name="A",
        regex="A",
        min_match=2,
        max_match=2,
    ), "Assert Min and max match follow each other if not set"
    assert RunbotExcludeWarning(name="A", regex="A", min_match=10, max_match=2) == RunbotExcludeWarning(
        name="A",
        regex="A",
        max_match=10,
        min_match=10,
    ), "Assert Min and max match follow each other if not set"
    assert RunbotExcludeWarning(name="A", regex="A", min_match=-1) == RunbotExcludeWarning(
        name="A",
        regex="A",
        max_match=1,
        min_match=0,
    ), "Assert if Min is -1 then this means 0 min match"
    assert RunbotExcludeWarning(name="A", regex="A", min_match=0) == RunbotExcludeWarning(
        name="A",
        regex="A",
        max_match=1,
        min_match=0,
    ), "Assert if Min is 0 then this means 0 min match"
    assert RunbotExcludeWarning(name="A", regex="A", max_match=0) == RunbotExcludeWarning(
        name="A",
        regex="A",
        max_match=0,
        min_match=0,
    ), "Assert if Max is 0 means exacly 0 match possible"
    assert RunbotExcludeWarning(name="A", regex="A", max_match=999) == RunbotExcludeWarning(
        name="A",
        regex="A",
        max_match=100,
        min_match=1,
    ), "Assert if Max can't be more than 100If you want more than 100, you should fix this logger :-)"



"""
Hook wrapper "result" utilities.
"""

from __future__ import annotations

from types import TracebackType
from typing import Callable
from typing import cast
from typing import final
from typing import Generic
from typing import Optional
from typing import Tuple
from typing import Type
from typing import TypeVar


_ExcInfo = Tuple[Type[BaseException], BaseException, Optional[TracebackType]]
ResultType = TypeVar("ResultType")


class HookCallError(Exception):
    """Hook was called incorrectly."""


@final
class Result(Generic[ResultType]):
    """An object used to inspect and set the result in a :ref:`hook wrapper
    <hookwrappers>`."""

    __slots__ = ("_result", "_exception")

    def __init__(
        self,
        result: ResultType | None,
        exception: BaseException | None,
    ) -> None:
        """:meta private:"""
        self._result = result
        self._exception = exception

    @property
    def excinfo(self) -> _ExcInfo | None:
        """:meta private:"""
        exc = self._exception
        if exc is None:
            return None
        else:
            return (type(exc), exc, exc.__traceback__)

    @property
    def exception(self) -> BaseException | None:
        """:meta private:"""
        return self._exception

    @classmethod
    def from_call(cls, func: Callable[[], ResultType]) -> Result[ResultType]:
        """:meta private:"""
        __tracebackhide__ = True
        result = exception = None
        try:
            result = func()
        except BaseException as exc:
            exception = exc
        return cls(result, exception)

    def force_result(self, result: ResultType) -> None:
        """Force the result(s) to ``result``.

        If the hook was marked as a ``firstresult`` a single value should
        be set, otherwise set a (modified) list of results. Any exceptions
        found during invocation will be deleted.

        This overrides any previous result or exception.
        """
        self._result = result
        self._exception = None

    def force_exception(self, exception: BaseException) -> None:
        """Force the result to fail with ``exception``.

        This overrides any previous result or exception.

        .. versionadded:: 1.1.0
        """
        self._result = None
        self._exception = exception

    def get_result(self) -> ResultType:
        """Get the result(s) for this hook call.

        If the hook was marked as a ``firstresult`` only a single value
        will be returned, otherwise a list of results.
        """
        __tracebackhide__ = True
        exc = self._exception
        if exc is None:
            return cast(ResultType, self._result)
        else:
            raise exc.with_traceback(exc.__traceback__)


# Historical name (pluggy<=1.2), kept for backward compatibility.
_Result = Result



from __future__ import annotations

import dataclasses
import logging
import subprocess
import typing
from pathlib import Path

import gitlab
from addons_installer import git_addons

if typing.TYPE_CHECKING:
    from addons_installer.addons_installer import OdooAddonsDef

    from .runbot_env import RunbotEnvironment

_logger = logging.getLogger("odoo_runbot")


@dataclasses.dataclass(frozen=True)
class AddonsToInstall:
    """Simple dataclass to hold the Odoo addons to install.

    With wich git branch is tried, the one chosen to clone,
    and if this addons need installation (git clone + pip install).

    Attributes:
        addon: The minimal definition of this addons
        branch_try: The complete liste of git branch (or tag) tried
        branch_to_use: The element in `branch_try` chossen to use on installation process
        to_install: This addons needs to be installed (False if already installed, or disabled)

    """

    addon: OdooAddonsDef
    branch_try: list[str] = dataclasses.field(default_factory=list)
    branch_to_use: str = dataclasses.field(default="")
    to_install: bool = True


class AddonsBranchResolver:
    """Contains the main logic to resolve branch (or tag) to use for each GitAddons found in the current environement.

    See `resolve_branch_addons_git` for more details
    Attributes:
        env: The current environement.
        addons_info: The addons to apply logic
    """

    def __init__(self, env: RunbotEnvironment, addons_info: set[OdooAddonsDef]) -> None:
        self.env = env
        self.addons_info = addons_info

    def _get_gitlab_api(self) -> gitlab.Gitlab | None:
        if not self.env.GITLAB_READ_API_TOKEN:
            return None

        gitlab_api = gitlab.Gitlab(self.env.CI_SERVER_URL, private_token=self.env.GITLAB_READ_API_TOKEN)
        try:
            gitlab_api.auth()
        except gitlab.GitlabAuthenticationError:
            gitlab_api = None
        return gitlab_api

    def resolve_branch_addons_git(self) -> list[AddonsToInstall]:
        """Main function for resolving branch addons.
        Important: Only support gitlab remote git server and it's API to resolve branch

        Get the list to try with `branch_to_try` function, then for each make an api call
        if this branch existe on the remote project then it's the choosen when and no futher try is perform.

        """
        results: list[AddonsToInstall] = []

        gitlab_api = self._get_gitlab_api()
        do_resolve_git = True
        if not gitlab_api:
            msg = "Disable auto find suitable branch"
            if not self.env.CI_API_V4_URL:
                msg += ", missing $CI_API_V4_URL"
            if not self.env.GITLAB_READ_API_TOKEN:
                msg += ", missing $GITLAB_READ_API_TOKEN"
            else:
                msg += ", $GITLAB_READ_API_TOKEN is set but probably not valid anymore. \nPlease contact your Tech Lead"
            _logger.error(msg)
            do_resolve_git = False

        for addon_info in self.addons_info:
            if not isinstance(addon_info, git_addons.GitOdooAddonsResult):
                _logger.warning(
                    "Addons %s isn't a git addons, so we can't change the branche",
                    addon_info.name,
                )
                results.append(AddonsToInstall(addon_info, ["[red] -"], "[red] -"))
                continue
            if Path(addon_info.addons_path).exists():
                branch_name = self._get_branch_name_of_path(addon_info)
                results.append(AddonsToInstall(addon_info, ["[red] already cloned"], "[green] " + branch_name))
                continue

            if not do_resolve_git:
                continue

            current_gl_project = gitlab_api.projects.get(self.env.CI_PROJECT_PATH)
            branch_to_try = self.branch_to_ty(addon_info, current_gl_project)
            remote_project = gitlab_api.projects.get(addon_info.git_path)
            branch_to_use = self.try_find_suitable_branch(remote_project, branch_to_try)
            addon_info.branch = branch_to_use
            if remote_project.visibility != "public":
                addon_info.protocole = "https"
                addon_info.format = git_addons.GitOdooAddons.FORMAT_GIT_CLONE["https"]
                addon_info.https_login = "gitlab-ci-token"
                addon_info.https_password = self.env.CI_DEPLOY_TOKEN or self.env.CI_JOB_TOKEN

            results.append(AddonsToInstall(addon_info, branch_to_try, "[green] " + branch_to_use))
        return results

    def _get_branch_name_of_path(self, addon_info: OdooAddonsDef) -> str:
        return subprocess.check_output(  # noqa: S603
            ["/usr/bin/git", "-C", addon_info.addons_path, "branch", "--show-current"],
        ).decode("utf-8")

    def branch_to_ty(
        self,
        addon: git_addons.GitOdooAddonsResult,
        project: gitlab.v4.objects.projects.Project,
    ) -> list[str]:
        """Return a list of branch names to try on a given project.
        The list is by default composed of:
        - The current branch name -> "CI_COMMIT_REF_NAME"
        - The target branch of the current merge request -> "CI_MERGE_REQUEST_TARGET_BRANCH_NAME"
        - The default branch of the current project
        - The branch for the addon, set with "ADDONS_GIT_<NAME>_BRANCH"
        - The current Odoo version (Last fallback)

        Args:
            addon: The git addons found in env to install.
            project: The curent GitlabProject on wich the runbot is executed (Value of $CI_PROJECT_PATH in env)

        Returns:
            The list of branch to try

        """
        # using `list of dict.fromkeys` to get a list of unique value without loosing the order of the list
        return [
            r
            for r in dict.fromkeys(
                [
                    self.env.CI_COMMIT_REF_NAME,
                    self.env.CI_MERGE_REQUEST_TARGET_BRANCH_NAME,
                    project.default_branch,
                    addon.branch,
                    self.env.ODOO_VERSION,
                ],
            )
            if r
        ]

    def try_find_suitable_branch(
        self,
        project: gitlab.v4.objects.projects.Project,
        branch_to_try_names: list[str],
    ) -> str:
        for branch_name in dict.fromkeys(branch_to_try_names):
            if not branch_name:
                continue
            try:
                return project.branches.get(branch_name).name
            except gitlab.GitlabGetError:
                _logger.warning("Can't fetch branch %s from Gitlab", branch_name)
        _logger.debug(
            "No suitable branche, fallback to default of project %s(%s)", project.name, project.default_branch
        )
        return project.default_branch



import logging

_logger = logging.getLogger(__name__)
try:
    from odoo import models

    from . import runbot_tester
except ImportError:
    _logger.error("Please install odoo to use runbot tester")



from odoo.tests import common


class TestStep4(common.TransactionCase):
    def test_filter(self):
        # Step3 - Filter2 will capture this line but not Filter1 (no same logger name)
        base_msg = "My Warning message - in %s"
        self.env["runbot.logger"].logger(None).warning(base_msg, "root")
        self.env["runbot.logger"].logger("Custom.logger").warning(base_msg, "custom")
        self.env["runbot.logger"].logger("odoo.addons.base").warning(base_msg, "base")



from __future__ import annotations

import json
import logging
import os
from pathlib import Path
from typing import List

import typer
from rich.console import Console, ConsoleOptions, Group, RenderResult
from rich.panel import Panel
from rich.table import Table
from typer import Typer
from typing_extensions import Annotated

from . import runbot_init
from .runbot_env import RunbotEnvironment, RunbotStepConfig, RunbotToolConfig, StepAction
from .runbot_run import StepRunner

app = Typer()

env: RunbotEnvironment = None


def rich_force_colors() -> bool:
    """In Gitlab CI runner there is no tty, and no color.

    This function force the color even if there is no TTY
    See Also:
        https://github.com/nf-core/tools/pull/760/files
        https://github.com/Textualize/rich/issues/343
    Returns:
        True if in CI/CD Job or if color is forced
    """
    return bool(
        os.getenv("CI")
        or os.getenv("GITHUB_ACTIONS")
        or os.getenv("FORCE_COLOR")
        or os.getenv("PY_COLORS")
        or os.getenv("COLORTERM") == "truecolor"
    )


console = Console(width=150, force_terminal=rich_force_colors())

_logger = logging.getLogger("odoo_runbot")


@app.callback()
def _callback(workdir: Path = None, *, verbose: bool = False) -> None:
    global env  # noqa: PLW0603
    env = RunbotEnvironment(dict(os.environ), workdir=workdir, verbose=verbose)
    env.setup_logging_for_runbot(console)
    env.print_info()
    if not env.check_ok():
        raise TypeError(300)


@app.command("init")
def init_runbot() -> None:
    """Init the current project to run test
    - Find external addons depedency, install them if needed (git clone + pip install)
    - Init database, and wait postgresql is ready
    - Create basic config file for Odoo using $ODOO_RC
    """
    current_project_key = f"ADDONS_LOCAL_{env.abs_curr_dir.name.upper()}"
    project_config = RunbotToolConfig.from_env(env)
    _logger.info(
        "Add current project %s=%s ? %s",
        current_project_key,
        str(env.abs_curr_dir),
        project_config.include_current_project,
    )
    if project_config.include_current_project:
        env.environ[current_project_key] = str(env.abs_curr_dir)
    info, addons = runbot_init.show_addons(env)
    console.print(info)
    runbot_init.install_addons(addons)
    console.print(runbot_init.init_database(env))
    console.print(runbot_init.init_config(env))


@app.command("run", help="""Run the steps after initializing the project""")
def run_runbot(
    steps: Annotated[List[str], typer.Option()] = None,
    only_action: Annotated[StepAction, typer.Option()] = None,
) -> None:
    """Run the steps after initializing the project

    You can filter the step you want to run with the `step_name` argument.

    warning:
        If no step is run, then mangono-runbot will exit with code 100.

    Args:
        steps:  The step to run, if None, are "all" then no filter is applied
        only_action:  Choose wich action to only run

    """
    step_names = set(steps if steps is not None else env.environ.get("RUNBOT_STEPS", "all").split(","))
    _logger.info("Running %s steps", list(step_names))
    if not Path(env.ODOO_RC).exists():
        _logger.error("[red] Please run `mangono-runbot init config` to create your odoo config file")
        raise typer.Abort

    project_config = RunbotToolConfig.from_env(env)
    steps_to_run = []

    for step in project_config.steps:
        if only_action is not None and step.action != only_action:
            _logger.debug("Filter step %s, only want action=%s", step.name, only_action.name)
            continue
        if step_names and not step_names.intersection({None, "all", step.name}) and step.action != StepAction.INSTALL:
            _logger.info("Skip %s", step.name)
            continue
        steps_to_run.append(step)

    step_runner = StepRunner(env)
    step_runner.setup_warning_filter(project_config.warning_filters)
    step_runner.setup_odoo()
    for step in steps_to_run:
        console.print(RichStep(step))
        rc = step_runner.execute(step)
        if rc:
            console.print(f"[red] Step {step.name} {bool_to_emoji(False)}")  # noqa: FBT003
        console.print(f"[green] Step {step.name} {bool_to_emoji(True)}")  # noqa: FBT003
        write_exit_code(step, return_code=rc, failfast=project_config.failfast)

    if not step_runner.has_run:
        write_exit_code(100)


def write_exit_code(step: RunbotStepConfig, *, return_code: int, failfast: bool) -> None:
    if env.in_ci:
        datas = {}
        exit_file = Path(env.result_path / "runbot-exit.json")
        exit_file.touch(exist_ok=True)
        if exit_file.stat().st_size > 0:
            datas = json.load(exit_file.open())
        datas[step.name] = return_code
        exit_file.write_text(json.dumps(datas))
        color = "red" if return_code else "green"
        console.print(f"[{color}] Write exit code '{return_code}' to {exit_file!s}")

    if failfast and return_code:
        rc_code = return_code
        if env.in_ci:
            rc_code = 0
        raise typer.Exit(rc_code)


def get_exit_codes() -> dict[str, int]:
    exit_file = Path(env.result_path / "runbot-exit.json")
    if not exit_file.exists():
        return {}
    return json.load(exit_file.open())


@app.command("report")
def report() -> None:
    project_config = RunbotToolConfig.from_env(env)
    rc_code = 0
    for step_name, rc in get_exit_codes().items():
        step = project_config.get_step(step_name)
        console.print(RichStep(step, "red" if rc > 0 else None))
        rc_code += rc
    step_runner = StepRunner(env)
    coverage = step_runner.get_coverage()
    coverage.load()
    coverage.report()
    coverage.xml_report(outfile=str(env.result_path / "coverage.xml"))
    coverage.html_report(directory=str(env.result_path / "coverage_html"))
    raise typer.Exit(rc_code)


def bool_to_emoji(v: bool) -> str:  # noqa: FBT001
    return ":heavy_check_mark:" if v else ":x:"


ASCII_ART_MANGONO = """
                          [medium_spring_green]      %    %             [/]
                          [medium_spring_green]    %%%   %%%   %%%%%%%  [/]
                          [medium_spring_green]  %%%%   %%%   %%%    %%%[/]
                          [medium_spring_green]%%%%     %%%   %%     %%%[/]
                          [medium_spring_green]  %%%%   %%%   %%%    %%%[/]
                          [medium_spring_green]    %%%   %%%   %%%%%%%  [/]
                          [medium_spring_green]      %    %             [/]

          @@@    @@@@   @@@@    @@@   @@   @@@@@@    @@@@@@   @@@   @@   @@@@@@    [medium_spring_green]%%%[/]
          @@@@   @@@@   @@@@    @@@@  @@  @@   @@@  @@@  @@@@ @@@@  @@  @@@  @@@   [medium_spring_green]  %%%[/]
          @@@@@ @@@@@  @@@ @@   @@ @@ @@ @@@ @@@@@@@@@    @@@ @@ @@ @@ @@@    @@@  [medium_spring_green]   %%%%[/]
          @@@ @@@ @@@ @@@@@@@@  @@  @@@@  @@    @@  @@@  @@@@ @@  @@@@  @@@  @@@   [medium_spring_green]  %%%[/]
          @@@ @@@ @@@ @@    @@@ @@   @@@   @@@@@@    @@@@@@   @@   @@@   @@@@@@    [medium_spring_green]%%%[/]

             @ @ @@ @  @@ @ @ @@@ @ @@  @@ @@@ @ @  [medium_spring_green]%%%%%%%%%[/]
                                                    [medium_spring_green]%%%%%%%%%[/]
"""


@app.command("diag")
def diag_print() -> None:
    console.print(ASCII_ART_MANGONO)
    from . import __about__ as about

    console.print(f"Version: {about.__version__}")
    console.print(f"Main Author: {about.__author__}({about.__author_email__})")
    console.print(f"Workdir: {env.abs_curr_dir}")
    console.print("Result (Test & Coverage): %s", env.result_path)
    t_warn = Table(
        "Name",
        "Action",
        "Message Filter",
        "Wanted Category",
        title="py.warnings Filters",
    )
    project_config = RunbotToolConfig.from_env(env)
    if project_config.warning_filters:
        for log_filter in project_config.warning_filters:
            t_warn.add_row(
                log_filter.name,
                log_filter.action,
                f"r'{log_filter.message}'",
                log_filter.category,
            )
    else:
        t_warn.add_row("[DEFAULT] No py.warnings allowed", "always", ".*", "Warnings", "*")
    console.print(t_warn)
    table = Table(
        "Step",
        "Module",
        "Run tests",
        "Activate Coverage",
        "Tags to test",
        "Logger filter",
        title="Steps",
    )
    for step in project_config.steps:
        table.add_row(
            step.name,
            (step.modules and ",".join(step.modules)) or str(step.modules),
            step.action.name,
            bool_to_emoji(step.action == StepAction.TESTS and step.coverage),
            ",".join(step.test_tags),
            ",".join([f.name for f in step.log_filters]),
        )
    console.print(table)

    for step in project_config.steps:
        console.print(RichStep(step))


class RichStep:
    def __init__(self, step: RunbotStepConfig, color: str | None = None) -> None:
        self.step = step
        self._color = color or "green" if self.step.action == StepAction.TESTS else "dodger_blue2"

    def __rich_console__(self, console: Console, options: ConsoleOptions) -> RenderResult:
        t_log = Table("Name", "Regex", "Match", "logger", title="Log Filters", width=146)
        if self.step.log_filters:
            for log_filter in self.step.log_filters:
                match_txt = f"Between {log_filter.min_match} and {log_filter.max_match}"
                if log_filter.min_match == log_filter.max_match:
                    match_txt = f"Exactly {log_filter.min_match}"
                t_log.add_row(
                    log_filter.name,
                    f"r'{log_filter.regex}'",
                    match_txt,
                    log_filter.logger + " (and all child logger)",
                )
        else:
            t_log.add_row("[DEFAULT] No log allowed", ".*", "Exactly 0", "odoo (and all child logger)")

        yield Panel(
            Group(
                f"Install : {self.step.modules}",
                f"Action : {self.step.action.name}",
                f"Activate Coverage: {bool_to_emoji(self.step.action == StepAction.TESTS and self.step.coverage)}",
                f"Allow warnings: {bool_to_emoji(self.step.allow_warnings)}",
                f"Test Tags: {self.step.test_tags}",
                t_log,
            ),
            style=self._color,
            title=self.step.name,
        )


if __name__ == "__main__":
    app()



from __future__ import annotations

import argparse
from typing import Any


def metadata_impl(
    *,
    called_by_app: bool,  # noqa: ARG001
    field: str,
    compact: bool,
) -> None:
    import json
    import os

    from hatchling.bridge.app import Application
    from hatchling.metadata.core import ProjectMetadata
    from hatchling.metadata.utils import resolve_metadata_fields
    from hatchling.plugin.manager import PluginManager

    app = Application()

    root = os.getcwd()
    plugin_manager = PluginManager()
    project_metadata = ProjectMetadata(root, plugin_manager)

    metadata = resolve_metadata_fields(project_metadata)
    if field:  # no cov
        if field not in metadata:
            app.abort(f'Unknown metadata field: {field}')
        elif field == 'readme':
            app.display(metadata[field]['text'])
        elif isinstance(metadata[field], str):
            app.display(metadata[field])
        else:
            app.display(json.dumps(metadata[field], indent=4))

        return

    for key, value in list(metadata.items()):
        if not value:
            metadata.pop(key)

    if compact:
        app.display(json.dumps(metadata, separators=(',', ':')))
    else:  # no cov
        app.display(json.dumps(metadata, indent=4))


def metadata_command(
    subparsers: argparse._SubParsersAction,
    defaults: Any,  # noqa: ARG001
) -> None:
    parser = subparsers.add_parser('metadata')
    parser.add_argument('field', nargs='?')
    parser.add_argument('-c', '--compact', action='store_true')
    parser.add_argument('--app', dest='called_by_app', action='store_true', help=argparse.SUPPRESS)
    parser.set_defaults(func=metadata_impl)



from __future__ import annotations

import dataclasses
import enum
import importlib
import logging
import os
import pathlib
import sys

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib
import typing

from rich.logging import RichHandler

if typing.TYPE_CHECKING:
    from rich.console import Console

_logger = logging.getLogger("odoo_runbot")
# prefix for all env variable to force interactive input

RUNBOT_PREFIX = "RUNBOT_"
SET_ODOO_PREFIX = "SET_ODOO_"


def _apply_default_if_none(data_class: object) -> None:
    if not dataclasses.is_dataclass(data_class):
        err = f"Only valid call for dataclasses {type(data_class)}"
        raise ValueError(err)
    # Loop through the fields
    for field in dataclasses.fields(data_class):
        # If there is a default and the value of the field is none we can assign a value
        if not isinstance(field.default, type(dataclasses.MISSING)) and getattr(data_class, field.name) is None:
            setattr(data_class, field.name, field.default)
        if not isinstance(field.default_factory, type(dataclasses.MISSING)) and getattr(data_class, field.name) is None:
            setattr(data_class, field.name, field.default_factory())


_LEVEL_MAPPING = {
    logging.DEBUG: "[blue]DEBUG[/blue]",
    logging.INFO: "[green]INFO[/green]",
    logging.WARNING: "[yellow]WARNING[/yellow]",
    logging.ERROR: "[red]ERROR[/red]",
    logging.CRITICAL: "[bold red]CRITICAL[/bold red]",
}


class RunbotEnvironment:
    """The complete env of the runbot"""

    UNIQUE_ID: str
    """Unique string id, based on Gitlab CI_JOB_ID and CI_NODE_INDEX in case of parallel"""
    ODOO_VERSION: str
    """
    The Odoo version to test, provided by the image where the runbot run.
    """
    GITLAB_READ_API_TOKEN: str
    """
    The token used to access the Gitlab API.
    This token is taken from `CI_JOB_TOKEN` or `PERSONAL_GITLAB_TOKEN` or `GITLAB_TOKEN` environment variable.
    In GitLab CI job be sure `CI_JOB_TOKEN` have the correct right. [https://docs.gitlab.com/ee/ci/jobs/ci_job_token.html]
    """

    def __init__(self, environ: dict[str, str], *, workdir: pathlib.Path | None = None, verbose: bool = False) -> None:
        self.environ = environ
        self.verbose = verbose or environ.get("RUNBOT_VERBOSE", False)
        self.CI_COMMIT_REF_NAME = environ.get("CI_COMMIT_REF_NAME")
        self.CI_MERGE_REQUEST_TARGET_BRANCH_NAME = environ.get("CI_MERGE_REQUEST_TARGET_BRANCH_NAME")
        self.CI_PROJECT_NAME = environ.get("CI_PROJECT_NAME")
        self.CI_JOB_TOKEN = environ.get("CI_JOB_TOKEN")
        self.CI_DEPLOY_TOKEN = environ.get("CI_DEPLOY_TOKEN")
        self.CI_PROJECT_PATH = environ.get("CI_PROJECT_PATH")
        self.CI_API_V4_URL = environ.get("CI_API_V4_URL")
        self.CI_SERVER_URL = environ.get("CI_SERVER_URL")
        self.ODOO_VERSION = str(environ.get("ODOO_VERSION"))
        self.UNIQUE_ID = "-".join(
            [
                "job",
                environ.get("CI_JOB_ID") or environ.get("RUNBOT_RANDOM_ID", "0"),
                environ.get("CI_NODE_INDEX", "1"),
            ],
        )
        self.GITLAB_READ_API_TOKEN = (
            environ.get("GITLAB_READ_API_TOKEN") or environ.get("CI_JOB_TOKEN") or environ.get("GITLAB_TOKEN")
        )
        self.in_ci = "CI" in os.environ
        self.abs_curr_dir = pathlib.Path.cwd().absolute().resolve()
        if workdir:
            self.abs_curr_dir = workdir.resolve().absolute()
        self.result_path = self.abs_curr_dir / "runbot_result"
        self.ODOO_RC = environ.get("ODOO_RC", str(self.abs_curr_dir / "odoo-config.ini"))
        self.environ["DATABASE"] = "_".join(["runbot_db", self.UNIQUE_ID.replace("-", "_")])
        self.environ["DB_NAME"] = self.environ["DATABASE"]
        try:
            self.result_path.mkdir(parents=True, exist_ok=True)
        except PermissionError:
            if not self.result_path.exists():
                _logger.warning("Can't create folder %s", str(self.result_path))

    def chdir(self, new_dir: pathlib.Path | None) -> None:
        if new_dir:
            os.chdir(new_dir)
            self.__dict__.pop("project_config", None)

    @property
    def CI_PROJECT_DIR(self) -> pathlib.Path:  # noqa: N802
        """Returns: The CI_PROJECT_DIR provide by Gitlab CI as a [pathlib.Path][] object."""
        return pathlib.Path(os.environ["CI_PROJECT_DIR"])

    def setup_logging_for_runbot(self, console: Console) -> None:
        level = logging.DEBUG if self.verbose else logging.INFO
        rich_handler = RichHandler(
            level, console=console, rich_tracebacks=True, enable_link_path=not self.environ.get("CI", False)
        )
        rich_handler.addFilter(logging.Filter("odoo_runbot"))
        _logger = logging.getLogger("odoo_runbot")
        _logger.addHandler(rich_handler)
        _logger.setLevel(level)

    def check_ok(self) -> bool:
        res = True
        if not self.abs_curr_dir.joinpath("pyproject.toml").exists():
            _logger.error("Workdir %s don't exist", self.abs_curr_dir)
            res = False
        elif not RunbotToolConfig.get_toml_data(self.abs_curr_dir.joinpath("pyproject.toml")):
            _logger.error("No `tool.runbot` config in %s", self.abs_curr_dir.joinpath("pyproject.toml"))
            res = False
        try:
            from odoo import release

            if release.version < "12.0":
                _logger.error("This runbot can't run the test from Odoo less than 12.0")
                res = False
        except ImportError:
            pass

        return res

    def print_info(self) -> None:
        _logger.info("Run in %s", str(self.abs_curr_dir))
        _logger.info("Test result in %s", str(self.result_path))
        if os.getenv("ODOO_DEPENDS"):
            _logger.error("Il ne faut plus utiliser `ODOO_DEPENDS` ni même la renseigner dans le runbot.")
            _logger.error("Il faut au lieu utiliser les variables `ADDONS_GIT_XXX` supportée par `addons-installer`")
        _logger.info("Odoo version: %s", self.ODOO_VERSION)
        _logger.info("Odoo config file: %s", self.ODOO_RC)


@dataclasses.dataclass()
class RunbotExcludeWarning:
    """Container for a regex to exclude in the log
    Attributes:
        name (str): A name for this regex.
            Allow to print if this regex found a warning to exclude
        min_match (int): Min occurrence of this warning in the log. The runbot should failed otherwaise
        max_match (int): Max occurrence of this warning in the log. The runbot should failed otherwaise
        regex (str): A regular expression to exclude
    """

    name: str
    regex: str
    logger: str = dataclasses.field(default="")
    min_match: int = dataclasses.field(default=1)
    max_match: int = dataclasses.field(default=1)

    def __post_init__(self) -> None:
        _apply_default_if_none(self)
        self.max_match = min(max(self.max_match, self.min_match), 100) if self.max_match > 0 else 0
        self.min_match = max(min(self.min_match, self.max_match), 0) if self.max_match > 0 else 0


@dataclasses.dataclass()
class RunbotPyWarningsFilter:
    """Dataclass storing wich py.warnings to filter"""

    name: str
    action: str
    message: str | None = dataclasses.field(default=None)
    category: str | None = dataclasses.field(default=None)

    def __post_init__(self) -> None:
        _apply_default_if_none(self)


class StepAction(enum.Enum):
    TESTS = "tests"
    INSTALL = "install"


@dataclasses.dataclass()
class RunbotStepConfig:
    """Contain the config for one warmup"""

    name: str
    modules: list[str]
    action: StepAction = dataclasses.field(default=StepAction.TESTS)
    test_tags: list[str] = dataclasses.field(default_factory=list)
    coverage: bool = dataclasses.field(default=True)
    log_filters: list[RunbotExcludeWarning] = dataclasses.field(default_factory=list)
    allow_warnings: bool = dataclasses.field(default=True)

    def __post_init__(self) -> None:
        _apply_default_if_none(self)


@dataclasses.dataclass()
class RunbotToolConfig:
    """The class containing all the config to run the tests.
    The data are read from the `pyproject.toml` and focus on the `tool.mangono.runbot` section
    """

    include_current_project: bool = dataclasses.field(default=True)
    allowed_warnings: list[str] = dataclasses.field(default_factory=list)
    """Allow all warning (log line with level WARNING, or `py.warnings`) from any python modules"""
    steps: list[RunbotStepConfig] = dataclasses.field(default_factory=list)
    """The config for the test phase (after warmup)"""
    pretty: bool = dataclasses.field(default=True)
    """Use color and pretty printing in log"""
    warning_filters: list[RunbotPyWarningsFilter] = dataclasses.field(default_factory=list)
    """"""
    failfast: bool = dataclasses.field(default=True)

    def __post_init__(self) -> None:
        _apply_default_if_none(self)

    def get_step(self, step_name: str) -> RunbotStepConfig:
        for step in self.steps:
            if step.name == step_name:
                return step
        msg = f"No such step: {step_name}"
        raise KeyError(msg)

    @classmethod
    def load_from_toml(cls, path: pathlib.Path) -> RunbotToolConfig:
        """Create a config from a TOML file path using tomllib or toml depending of the python version.

        Args:
            path:

        Returns:

        """
        return cls.load_from_toml_data(cls.get_toml_data(path))

    @classmethod
    def get_toml_data(cls, path: pathlib.Path) -> dict[str, typing.Any]:
        """Create a config from a TOML file path using tomllib or toml depending of the python version.

        Args:
            path:

        Returns:

        """
        with path.open(mode="rb") as pyproject_toml:
            data = tomllib.load(pyproject_toml)
        return data.get("tool", {}).get("runbot", {})

    @classmethod
    def from_env(cls, env: RunbotEnvironment) -> RunbotToolConfig:
        return cls.load_from_toml(env.abs_curr_dir.joinpath("pyproject.toml"))

    @classmethod
    def load_from_toml_data(cls, runbot_data: dict) -> RunbotToolConfig:
        """Convert the Toml data to a RunbotToolConfig object.

        Args:
            runbot_data: All the data for the sub kley of `runbot.tool`

        Returns: A `RunbotToolConfig` object

        """
        version_log_filter = log_filter_by_odoo_version()
        global_log_filter = cls._get_log_filters(version_log_filter, runbot_data)
        global_pywarnings = warning_filter_by_odoo_version() + [
            RunbotPyWarningsFilter(
                name=py_filter_data.get("name", f"Global PyWarnings Filter {_idx}"),
                action=py_filter_data.get("action"),
                message=py_filter_data.get("message"),
                category=py_filter_data.get("category"),
            )
            for _idx, py_filter_data in enumerate(runbot_data.get("pywarnings-filter", []))
        ]

        cls._inject_default_step(runbot_data)

        steps = []
        for _idx_step, (step_name, step_data) in enumerate(runbot_data.get("step").items()):
            log_filter = cls._get_log_filters(global_log_filter, step_data, step_name=step_name)
            action = cls._get_action(step_name, step_data)
            step_obj = RunbotStepConfig(
                name=step_name,
                allow_warnings=step_data.get("allow-warnings", runbot_data.get("allow-warnings")),
                coverage=action == StepAction.TESTS and step_data.get("coverage", runbot_data.get("coverage")),
                modules=step_data.get("modules", runbot_data.get("modules")),
                action=action,
                test_tags=step_data.get("test-tags"),
                log_filters=log_filter,
            )
            steps.append(step_obj)

        return cls(
            include_current_project=runbot_data.get("include-current-project"),
            pretty=True,
            failfast=runbot_data.get("failfast"),
            steps=steps,
            warning_filters=global_pywarnings,
        )

    @classmethod
    def _inject_default_step(cls, runbot_data: dict[str, typing.Any]) -> None:
        if not runbot_data.get("step"):
            runbot_data["step"] = {
                "default": {
                    "coverage": runbot_data.get("coverage"),
                    "unittest-output": runbot_data.get("unittest-output"),
                    "modules": runbot_data.get("modules"),
                    "action": StepAction.TESTS.name,
                    "test-tags": [],
                    "log-filter": [],
                }
            }

    @classmethod
    def _get_log_filters(
        cls,
        global_log_filter: list[RunbotExcludeWarning],
        step_data: dict[str, typing.Any],
        step_name: str | None = None,
    ) -> list[RunbotExcludeWarning]:
        log_filter = (global_log_filter and global_log_filter[:]) or []
        for _idx_log_filter, data in enumerate(step_data.get("log-filters", []), start=len(log_filter) + 1):
            default_name = f"All Steps - Logger Filter {_idx_log_filter}"
            if step_name:
                default_name = f"Step {step_name} - Logger Filter {_idx_log_filter}"
            _data = data
            if isinstance(data, str):
                _data = {"regex": data}
            log_filter.append(
                RunbotExcludeWarning(
                    name=_data.get("name", default_name),
                    regex=_data["regex"],
                    logger=_data.get("logger"),
                    min_match=_data.get("min-match"),
                    max_match=_data.get("max-match"),
                ),
            )
        return log_filter

    @staticmethod
    def _get_action(step_name: str, step_data: dict[str, typing.Any]) -> StepAction:
        config_value = step_data.get("action", step_name)
        if config_value in ("install", "warmup"):
            return StepAction.INSTALL
        return StepAction.TESTS


def log_filter_by_odoo_version() -> list[RunbotExcludeWarning]:
    """Returns a list of `RunbotExcludeWarning` objects predefined by odoo version
    This list avoid duplication accross project to filter odoo log warning.

    Returns: The list of `RunbotExcludeWarning` objects for the current odoo version or an empty list

    """
    if importlib.util.find_spec("odoo"):
        return [
            RunbotExcludeWarning(
                name="Default - unaccent not loadable",
                logger="odoo.modules.registry",
                regex=r".*no unaccent\(\) function was found in database.*",
                min_match=0,
                max_match=99,
            ),
        ]

    return []


def warning_filter_by_odoo_version() -> list[RunbotPyWarningsFilter]:
    """Returns a list of `RunbotPyWarningsFilter` objects predefined by odoo version
    This list avoid duplication accross project to filter odoo log warning.

    Returns: The list of `RunbotPyWarningsFilter` objects for the current odoo version or an empty list

    """
    if importlib.util.find_spec("odoo"):
        return [
            RunbotPyWarningsFilter(
                name="[Default] Exclude error in reportlab/pdfbase",
                action="ignore",
                category=SyntaxWarning.__name__,
                message='.*"is" with a literal. Did you mean.*',
            ),
            RunbotPyWarningsFilter(
                name="[Default] Exclude error in vobject/base.py",
                action="ignore",
                category=SyntaxWarning.__name__,
                message=".*invalid escape sequence*",
            ),
        ]

    return []



import logging

from odoo.tests import common


class TestStep3(common.TransactionCase):
    def test_filter(self):
        """
        Test the filter of the Step3.
        This filter should capturing the log with the default name of the model `runbot.logger`
        -> `odoo.addons.runbot_tester.models.runbot_tester` it's a child of `odoo.addons.runbot_tester`

        """
        # Step3 - Filter1 should capture this line
        self.env["runbot.logger"].logger().warning("My Warning Filter1")
        print(logging.root.handlers)
        print(logging.root.handlers[0].filters)
        print(logging.root.handlers[1].filters)
        # In this level this should be ok
        # Even if the message match
        self.env["runbot.logger"].logger().info("My Warning Filter1")



"""Contains all the code used to assert Logger and warnings a propely catched."""

from __future__ import annotations

import logging
import types
import typing
import unittest

from xmlrunner import XMLTestRunner
from xmlrunner.result import _XMLTestResult

if typing.TYPE_CHECKING:
    import pathlib
    from warnings import WarningMessage

    from .runbot_run_logging import ExcludeWarningFilter

_logger = logging.getLogger("odoo_runbot")


class XmlResultSilent(_XMLTestResult):
    def _is_relevant_tb_level(self, tb):  # noqa: ANN001 ARG002 ANN202
        return True


def get_test_runner(output_dir: pathlib.Path | None = None) -> unittest.TextTestRunner:
    """Retun a default Runner.

    If an `output_dir` is filed, then `xmlrunner.XMLTestRunner` is returned
    Args:
        output_dir: The path where the test resul;t should be stored

    Returns: The class to run unittest tests

    """
    if output_dir and output_dir.exists():
        return XMLTestRunner(output=str(output_dir.resolve()), failfast=False, resultclass=XmlResultSilent, verbosity=2)
    return unittest.TextTestRunner(failfast=False, resultclass=XmlResultSilent, verbosity=2)


def execute_test_after_odoo(
    dynamic_class_name: str,
    log_filters: list[ExcludeWarningFilter] | None = None,
    warning_message: list[WarningMessage] | None = None,
    *,
    test_runner: unittest.TextTestRunner | None = None,
) -> unittest.result.TestResult:
    """Generate 1 [unittest.TestCase][] for each filter.

    If `test_runner` is `None` then [unittest.TextTestRunner][] is used.

    Note:
        If `filter_to_test` is `None` then A default succeed TestResult is returned.

    Args:
        dynamic_class_name: The name of the class to generate, No sneedTo start with "Test"
        log_filters: All the Log filter to add in the test suite
        warning_message: All the warning message catch, if any then the suite will failed
        test_runner: your custom runner, or [unittest.TextTestRunner][]

    Returns:
        The result of the test runner with your runner if set.

    """
    _logger.info("Execute Logger Filter test suite for %s filters", (log_filters and len(log_filters)) or 0)
    test_runner = test_runner or get_test_runner(None)
    if not log_filters and not warning_message:
        return unittest.result.TestResult()
    test_class = types.new_class(f"Test{_slug_str(dynamic_class_name)}", (_RunbotLoggerUnitTest,), {})
    generated_test_func = []
    suite = unittest.TestSuite()
    for log_filter in log_filters:
        dynamic_name = f"test_{_slug_str(log_filter.exclude.name)}"
        generated_test_func.append((dynamic_name, log_filter))
        setattr(test_class, dynamic_name, test_class._template_logger_filter)  # noqa: SLF001
        suite.addTest(test_class(dynamic_name, logger_filter=log_filter, warning_msg=None))
    if warning_message:
        test_class.test_warnings = test_class._template_test_warning_filter  # noqa: SLF001
        generated_test_func.append("test_warnings")
        suite.addTest(test_class("test_warnings", logger_filter=None, warning_msg=warning_message))

    _logger.debug("Test suite created, run it with %s", type(test_runner).__name__)
    return test_runner.run(suite)


def _slug_str(value: str) -> str:
    return "".join(d.capitalize() for d in value.split()).replace(" ", "").replace("-", "").lower()


class _RunbotLoggerUnitTest(unittest.TestCase):
    def __init__(
        self, method_name: str, logger_filter: ExcludeWarningFilter, warning_msg: list[WarningMessage]
    ) -> None:
        super().__init__(method_name)
        self.logger_filter = logger_filter
        self._testMethodDoc = self._make_description()
        self.warning_filter = warning_msg
        self._testMethodDoc = ""
        if logging.root.handlers:
            self.format_log = logging.root.handlers[0].format
        else:
            self.format_log = logging.Formatter.format

    def _template_logger_filter(self) -> None:
        self.assertTrue(self.logger_filter)  # noqa: PT009
        if self.logger_filter.success:
            self.assertTrue(self.logger_filter.success)  # noqa: PT009
            return

        fail_msg = [
            f"{self.logger_filter.exclude.name} Failed for logger '{self.logger_filter.exclude.logger or 'all'}'",
            f"Expected: {self.logger_filter.exclude.min_match} "
            f"<= len(match_log_lines) <= {self.logger_filter.exclude.max_match}",
            f"Found {len(self.logger_filter.log_match)} "
            f"log line who match the regex r'{self.logger_filter.regex.pattern}'",
        ]
        if self.logger_filter.log_match:
            fail_msg.append("Log line content :")
        fail_msg.extend(["\t> " + self.format_log(log_record) for log_record in self.logger_filter.log_match])
        self.fail("\n".join(fail_msg))

    def _make_description(self) -> str:
        if self.logger_filter:
            return (
                f"{self.logger_filter.exclude.name} Regex Matcher "
                f"for logger '{self.logger_filter.exclude.logger or 'all'}'"
            )
        return ""

    def _template_test_warning_filter(self) -> None:
        self.assertTrue(self.warning_filter)  # noqa: PT009
        fail_msg = ["Some py.warnings found"] + ["\t> " + str(m) for m in self.warning_filter]
        self.fail(msg="\n".join(fail_msg))


class _RunbotPyWarningUnitTest(unittest.TestCase):
    def __init__(self, warning_msg: list[WarningMessage]) -> None:
        super().__init__("test_warning_filter")
        self.warning_filter = warning_msg
        self._testMethodDoc = ""



from __future__ import annotations

import os
from abc import ABC, abstractmethod
from functools import cached_property


class VersionSchemeInterface(ABC):  # no cov
    """
    Example usage:

    ```python tab="plugin.py"
    from hatchling.version.scheme.plugin.interface import VersionSchemeInterface


    class SpecialVersionScheme(VersionSchemeInterface):
        PLUGIN_NAME = 'special'
        ...
    ```

    ```python tab="hooks.py"
    from hatchling.plugin import hookimpl

    from .plugin import SpecialVersionScheme


    @hookimpl
    def hatch_register_version_scheme():
        return SpecialVersionScheme
    ```
    """

    PLUGIN_NAME = ''
    """The name used for selection."""

    def __init__(self, root: str, config: dict) -> None:
        self.__root = root
        self.__config = config

    @property
    def root(self) -> str:
        """
        The root of the project tree as a string.
        """
        return self.__root

    @property
    def config(self) -> dict:
        """
        ```toml config-example
        [tool.hatch.version]
        ```
        """
        return self.__config

    @cached_property
    def validate_bump(self) -> bool:
        """
        This is the value of the `validate-bump` option, with the `HATCH_VERSION_VALIDATE_BUMP`
        environment variable taking precedence. Validation is enabled by default.

        ```toml config-example
        [tool.hatch.version]
        validate-bump = true
        ```
        """
        from hatchling.utils.constants import VersionEnvVars

        if VersionEnvVars.VALIDATE_BUMP in os.environ:
            return os.environ[VersionEnvVars.VALIDATE_BUMP] not in {'false', '0'}

        validate_bump = self.config.get('validate-bump', True)
        if not isinstance(validate_bump, bool):
            message = 'option `validate-bump` must be a boolean'
            raise TypeError(message)

        return validate_bump

    @abstractmethod
    def update(self, desired_version: str, original_version: str, version_data: dict) -> str:
        """
        This should return a normalized form of the desired version. If the
        [validate_bump](reference.md#hatchling.version.scheme.plugin.interface.VersionSchemeInterface.validate_bump)
        property is `True`, this method should also verify that the version is higher than the original version.
        """



from odoo.tests import common


class TestStep6(common.TransactionCase):
    def test_simple_filter(self):
        self.env["runbot.logger"].logger("odoo.addons.base").warning("message for %s", "step6")



{
    "name": "Runbot Odoo Module",
    "version": "1.3",
    "category": "Hidden",
    "description": "Sample Odoo runbot tester",
    "data": [],
    "demo": [],
    "assets": {},
    "installable": True,
    "auto_install": False,
    "license": "LGPL-3",
}



import logging

from odoo_runbot import runbot_run_logging as log_handler
from odoo_runbot.runbot_env import RunbotExcludeWarning, RunbotStepConfig
from odoo_runbot.runbot_run_logging import ExcludeWarningFilter, RunbotWarningWatcherHandler
from odoo_runbot.runbot_run_test import execute_test_after_odoo

_logger = logging.getLogger("odoo")


def teardown_function():
    """On teardown, remove all runbot logging handlers"""
    for handler in logging.root.handlers[:]:
        if isinstance(handler, RunbotWarningWatcherHandler):
            logging.root.removeHandler(handler)


def test_log_setup_1_step():
    warn_rule = RunbotExcludeWarning(name="MyRegex", min_match=2, max_match=2, regex=".*This is a warn.*")
    step = RunbotStepConfig(
        name="test",
        modules=["module1"],
        log_filters=[warn_rule],
    )
    assert log_handler.get_handler() not in logging.root.handlers, "The handler is not activated"
    with log_handler.start_warning_log_watcher(step) as log_filters:
        global_handler = log_handler.get_handler()
        assert log_handler.get_handler() in logging.root.handlers, "Runbot Handler activated"
        assert global_handler is not None, "Should be initialized after `start_warning_log_watcher`"
        assert len(global_handler.record_matchers()) == 1, "One warnings rule in step eq 1 filter in log handler"
        assert isinstance(global_handler.record_matchers()[0], ExcludeWarningFilter), (
            "contains only ExcludeWarningFilter"
        )
        assert global_handler.record_matchers()[0].exclude == warn_rule

    log_filters()
    assert not log_handler.get_handler().record_matchers(), "After stop no more warnings rule registered"


def test_log_setup_2_step():
    warn_rule1 = RunbotExcludeWarning(name="warn_rule1", regex=".*This is a warn.*")
    step1 = RunbotStepConfig(
        name="step1",
        modules=["module1"],
        log_filters=[warn_rule1],
    )
    warn_rule2_1 = RunbotExcludeWarning(name="warn_rule2_1", regex=".*This is a warn.*")
    warn_rule2_2 = RunbotExcludeWarning(name="warn_rule2_2", regex=".*This is a warn2.*")
    step2 = RunbotStepConfig(
        name="step2",
        modules=["module1"],
        log_filters=[warn_rule2_1, warn_rule2_2],
    )

    assert log_handler.get_handler() not in logging.root.handlers, "The handler is not activated"
    with log_handler.start_warning_log_watcher(step2) as extractor2:
        assert log_handler.get_handler() in logging.root.handlers, "Runbot Handler step1 activated"
        global_handler = log_handler.get_handler()
        assert len(global_handler.record_matchers()) == 2, "2 filter in step 2"
        with log_handler.start_warning_log_watcher(step1) as extractor1:
            assert len(global_handler.record_matchers()) == 3, "1 filter in step 1 + 2 filters in step2"
            assert all(isinstance(_filter, ExcludeWarningFilter) for _filter in global_handler.record_matchers()), (
                "contains only ExcludeWarningFilter"
            )
            msg1 = "This is a warning. - captured by warn_rule2_1 filter first"
            _logger.warning(msg1)

        filter_warnings = extractor1()
        assert len(filter_warnings) == 1, "1 Filter in step1"
        assert filter_warnings[0].exclude == warn_rule1
        assert len(filter_warnings[0].log_match) == 0, "Catch 0 log line, warn_rule2_1 catch it first"
        assert len(log_handler.get_handler().record_matchers()) == 2, "Still have the 2 filter from step2"
        msg2 = "This is a warning. - captured by warn_rule2_1 filter"
        msg3 = "This is a warn2 - captured by warn_rule2_1 and warn_rule2_2 filters"
        msg4 = "This is a warn2 - not caputred, incorect log level"
        _logger.warning(msg2)
        _logger.warning(msg3)

        _logger.debug(msg4)
        _logger.info(msg4)

    filter_warnings2 = extractor2()
    assert len(filter_warnings2) == 3, "2 Filter in step2 + 1 for the default"
    assert filter_warnings2[0].exclude == warn_rule2_1
    assert filter_warnings2[1].exclude == warn_rule2_2
    assert filter_warnings2[0].log_match.pop(0).getMessage() == msg1
    assert filter_warnings2[0].log_match.pop(0).getMessage() == msg2
    assert filter_warnings2[0].log_match.pop(0).getMessage() == msg3
    assert not filter_warnings2[0].log_match, "More than 3 log catched ? " + str(filter_warnings2[0].log_match)

    assert not filter_warnings2[1].log_match, "No message to catch (all catch by previous filter" + str(
        filter_warnings2[1].log_match,
    )


def test_log_warnings_as_test():
    """I only wan't one line of log with message Containing "This is a warn"
    If 0 -> Failed
    if 1 -> success
    if 2 -> failed
    """
    step = RunbotStepConfig(
        name="test",
        modules=["module1"],
        log_filters=[
            RunbotExcludeWarning(
                name="MyRegex",
                min_match=2,
                max_match=2,
                regex=".*This is a warn.*",
            ),
        ],
    )
    logger = logging.getLogger("odoo")
    sub_logger = logging.getLogger("odoo.sql")
    with log_handler.start_warning_log_watcher(step) as extractor:
        logger.warning("This is a warning line captured")
        sub_logger.warning("This is a warning line captured (second)")
    filters = extractor()
    test_result = execute_test_after_odoo("Fake", filters)
    assert test_result.wasSuccessful(), "Should catch the 2 log line above"
    assert test_result.testsRun == 2, "My Step contains 1 warnings, and the default catch all"


def test_start_and_stop_watching():
    """I only want one line of log with message Containing "This is a warn"
    If 0 -> Failed
    if 1 -> success
    if 2 -> failed
    """
    step = RunbotStepConfig(
        name="test",
        modules=["module1"],
        log_filters=[
            RunbotExcludeWarning(
                name="MyRegex",
                logger="odoo.sql",
                min_match=1,
                max_match=1,
                regex=".*This is a warn.*",
            ),
        ],
    )
    assert log_handler.get_handler() not in logging.root.handlers, "The handler is not activated"
    with log_handler.start_warning_log_watcher(step) as extractor:
        handler_post_start = log_handler.get_handler()
        assert handler_post_start is not None, "After start, the handler is registered"
        assert len(handler_post_start.record_matchers()) == 1
    extractor()
    assert not log_handler.get_handler().filters, "After stop, no filter is activated"


def test_no_rule_default_catcher():
    "Create a step without filter, then all warning logs are catch and this should created a final failed test result"
    step = RunbotStepConfig(
        name="test",
        modules=["module1"],
        log_filters=[],
    )
    assert log_handler.get_handler() not in logging.root.handlers, "The handler is not activated"
    with log_handler.start_warning_log_watcher(step) as extractor:
        handler_post_start = log_handler.get_handler()
        assert handler_post_start is not None, "After start, the handler is registered"
        assert not handler_post_start.filters, "No filter"

        logging.getLogger("my_logger").warning("Catch this log warning")

    log_filter = extractor()
    result = execute_test_after_odoo("Fake", log_filter)
    assert not result.wasSuccessful(), "Should catch the 2 log line above"


def test_no_catch_rule_default_catcher():
    "Create a step without filter, then all warning logs are catch and this should created a final failed test result"
    step = RunbotStepConfig(
        name="test",
        modules=["module1"],
        log_filters=[
            RunbotExcludeWarning(
                name="MyRegex",
                logger="odoo.sql",
                min_match=1,
                max_match=1,
                regex=".*This is a warn.*",
            ),
        ],
    )
    assert log_handler.get_handler() not in logging.root.handlers, "The handler is not activated"
    with log_handler.start_warning_log_watcher(step) as extractor:
        logging.getLogger("my_logger").warning("No catched warning")

    log_filter = extractor()
    result = execute_test_after_odoo("Fake", log_filter)
    assert not result.wasSuccessful(), "Should catch the 2 log line above"


def test_no_catch_only_default_catcher():
    """Create a step without filter.

    then all warning logs are catch and this should created a final failed test result
    """
    step = RunbotStepConfig(
        name="test",
        modules=["module1"],
        log_filters=[],
    )
    assert log_handler.get_handler() not in logging.root.handlers, "The handler is not activated"
    with log_handler.start_warning_log_watcher(step) as extractor:
        logging.getLogger("my_logger").warning("No catched warning")
    log_filter = extractor()
    result = execute_test_after_odoo("Fake", log_filter)
    assert not result.wasSuccessful(), "Should catch the 2 log line above"


def test_catch_critical():
    step = RunbotStepConfig(
        name="test",
        modules=["module1"],
        log_filters=[
            RunbotExcludeWarning(
                name="MyRegex",
                regex=".*This is a warn.*",
                max_match=0,
            ),
        ],
    )
    assert log_handler.get_handler() not in logging.root.handlers, "The handler is not activated"
    with log_handler.start_warning_log_watcher(step) as extractor:
        logging.getLogger("my_logger").critical("critical %s", "msg")
        logging.getLogger("my_logger").error("error %s", "msg")
        logging.getLogger("my_logger").exception("exception %s", "msg")
        logging.getLogger("my_logger").warning("warning %s", "msg")
        logging.getLogger("my_logger").debug("debug %s", "msg")
        logging.getLogger("my_logger").info("info %s", "msg")

    log_filter = extractor()
    result = execute_test_after_odoo("Fake", log_filter)
    assert not result.wasSuccessful(), "Should catch the 2 log line above"
    assert result.testsRun == 2, "Default catcher + rule"
    assert len(result.failures) == 1
    assert "AssertionError: Runbot Warning catch all no filter Failed" in result.failures[0][1], (
        "Assert it's the default catcher, not the rule"
    )


def test_match_msg_with_args():
    log_filter = RunbotExcludeWarning(
        name="MyRegex",
        regex=".*A msg with args value=toto.*",
    )
    step = RunbotStepConfig(
        name="test",
        modules=["module1"],
        log_filters=[log_filter],
    )
    with log_handler.start_warning_log_watcher(step) as extractor:
        msg1 = "A msg with args value=%s"
        _logger.warning(msg1, "toto")  # Catch by my step filter
        _logger.warning(msg1, "tutu")  # Catch by global filter

    filter_warnings = extractor()
    assert len(filter_warnings) == 2, "1 Filter in step1 + Catch all"
    assert filter_warnings[0].exclude == log_filter
    assert bool(filter_warnings[0].log_match), "Should contains the msg with 'toto'"

    assert filter_warnings[0].log_match.pop(0).getMessage() == "A msg with args value=toto"
    assert filter_warnings[1].log_match.pop(0).getMessage() == "A msg with args value=tutu"
    assert not filter_warnings[0].log_match, "More than 2 log catched ? " + str(filter_warnings[0].log_match)
